#!/usr/bin/python3

from GNOMEAcquirerLibs.DataAcquire import *
from GNOMEAcquirerLibs.AcquisitionConfig import *
from GNOMEAcquirerLibs.DataAnalysis import *
from GNOMEAcquirerLibs.Meta import *
from GNOMEAcquirerLibs.ViewLog import *
import numpy as np
import h5py
import os
import sys
import tempfile
import xml.etree.ElementTree as ET
import socket
import urllib
import subprocess
import errno
from math import log
import traceback
import re
import gnomedata as gd
import sympy
import math

# constants

ConfigAttribute_Prefix = "att_"
ConfigAttribute_NameSuffix = "_name"
ConfigAttribute_ValueSuffix = "_value"
ConfigAttribute_TypeSuffix = "_type"
ConfigAttribute_ValueLockedSuffix = "_valueLocked"
ConfigAttribute_LockedSuffix = "_locked"
ConfigOption_Prefix             = "opt_"
ConfigOptionName_SelectedDevice     = "SelectedDevice"
ConfigOptionName_SelectedChannel    = "SelectedChannel"
ConfigOptionName_ChannelDatatype    = "SelectedDataType"
# ConfigOptionName_ChConversionFactor = "ConversionFactor"
# ConfigOptionName_ChOffset = "ChannelOffset"
ConfigOptionName_DataPath = "DataPath"
ConfigOptionName_StationName = "StationName"
ConfigOptionName_PortName = "SelectedDevice"

AcquisitionTypeOption_COM  = "Acquisition over COM"
AcquisitionTypeOption_Fake = "Fake acquisition"
AcquisitionTypeOption_File = "Acquisition from files..."

StationType_MagneticField = "Magnetic field sensor"
StationType_AtomicClock   = "Atomic Clock(s) station"

# fix window icon not showing up in the taskbar for ViewLog()
if sys.platform.startswith('win'):
    import ctypes

    myappid = 'budker.gnomeacquirer.subproduct.version'  # arbitrary string
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)


# a function that takes an attribute string, and returns it replaced by values when necessary according
# to pre-defined keywords

# a function that doesn't require data information for testing initial values of the attributes
def ApplyAttributeReplacement_tester(attrStr):
    TimeV = datetime.datetime.utcnow()
    attrStrReplaced = copy.deepcopy(attrStr)
    StartTime    = TimeV.strftime(DataBatch.f_timeFormat)
    EndTime      = TimeV.strftime(DataBatch.f_timeFormat)
    Date         = TimeV.strftime(DataBatch.f_dateFormat)
    SR           = "100"
    Weeknum      = "12345"
    Longitude    = "1.23"
    Latitude     = "3.45"
    Altitude     = "6.78"
    IntTemp      = "14.15"
    ExtTemp      = "12.13"
    ChRange      = "+/-1.25[V]"
    MissingPoints = "2"

    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_StartTime,StartTime)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_EndTime,EndTime)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_Date,Date)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_SamplingRate,SR)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_SamplingRate_SanityChannel,SR)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_WeekNumber,Weeknum)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_Altitude,Altitude)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_Latitude,Latitude)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_Longitude,Longitude)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_InternalTemperature,IntTemp)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_ExternalTemperature,ExtTemp)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_MissingPoints,MissingPoints)

    return attrStrReplaced

UpdateFileLink = "https://budker.uni-mainz.de//download/GNOMEAcquirer/updates.xml"


def human_readable_bytes(x):
    # hybrid of http://stackoverflow.com/a/10171475/2595465
    #     with http://stackoverflow.com/a/5414105/2595465
    if x <= 0: return '0 B'
    magnitude = math.floor(log(abs(x),10.24))
    if magnitude > 16:
        format_str = '%iP'
        denominator_mag = 15
    else:
        float_fmt = '%2.1f' if magnitude % 3 == 1 else '%1.2f'
        illion = math.floor(magnitude / 3)
        format_str = float_fmt + ['B', ' KB', ' MB', ' GB', ' TB', ' PB'][illion]
        val = (x * 1.0 / (1024 ** illion))
        val_str = (format_str % val).lstrip('0')
        if len(val_str) > 0 and val_str.startswith("."):
            val_str = "0" + val_str
    return val_str


def CheckForUpdates():
    # if not hasattr(sys, 'frozen'):
    #     return {"success": False, "msg": "Non frozen (non-executable) apps shall not upgraded from here."}

    try:
        with urllib.request.urlopen(UpdateFileLink) as f:
            print("Check for newer versions...")
            tree = ET.parse(f)
            root = tree.getroot()
            versionTree = root.find('version')
            downloadLinkTree = root.find('softwareLink')
            print("Current version: " + SOFTWARE_VERSION)
            print("Latest version: " + versionTree.text)
            if not IsVersionNewer(versionTree.text):
                return {'newerAvailable': False, 'url': '', 'version': versionTree.text, "success": True}
            else:
                return {'newerAvailable': True, 'url': downloadLinkTree.text,
                        'version': versionTree.text, "success": True}
    except Exception as e:
        traceback.print_exc()
        print("Error while trying to get update information. Exception says: " + str(e))
        return {"success": False, "msg": str(e)}


def IsVersionNewer(version):
    isBeta = (len(SOFTWARE_VERSION.split(".")) > 3)
    currentSoftwareVersion = ".".join(SOFTWARE_VERSION.split(".")[0:3]) #fourth part is for beta versions
    currentVersionSplit = list(map(int,currentSoftwareVersion.split(".")))
    newVersionSplit = list(map(int,version.split(".")))

    if isBeta and currentVersionSplit == newVersionSplit: #if the current version is beta and the new version is not beta, then take it
        return True

    # print(currentVersionSplit)
    # print(newVersionSplit)
    if len(newVersionSplit) != 3:
        raise Exception("Error while parsing retrieved version number. The version value " + version + " is not valid.")

    if int(newVersionSplit[0]) > int(currentVersionSplit[0]):
        return True
    elif int(newVersionSplit[0]) < int(currentVersionSplit[0]):
        return False

    if int(newVersionSplit[1]) > int(currentVersionSplit[1]):
        return True
    if int(newVersionSplit[1]) < int(currentVersionSplit[1]):
        return False

    elif int(newVersionSplit[2]) > int(currentVersionSplit[2]):
        return True
    elif int(newVersionSplit[2]) < int(currentVersionSplit[2]):
        return False

    return False


def clean(item):
    """Clean up the memory by closing and deleting the item if possible."""
    if isinstance(item, list) or isinstance(item, dict):
        for _ in range(len(item)):
            clean(list(item).pop())
    else:
        try:
            item.close()
        except: # deleted or no close method
            try:
                item.deleteLater()
            except: # deleted or no deleteLater method
                pass


Tooltip_StationName      = "Station name (preferably GNOME synchronizer username) to be inserted as a prefix to file names; e.g., berkeley01"
Tooltip_DataPath         = "Path to the folder where acquired data will be stored. This is the same folder that has to be chosen for the synchronizer."
Tooltip_AddDataset       = "Add a dataset to the saved HDF5 file. The dataset contains data that is acquired from a channel in the acquisition device, " \
                           "in addition to some attributes (header information) that is contained in the table shown in the dataset tab."
Tooltip_RemoveDataset    = "Remove an existing dataset"
Tooltip_AcquisitionType  = "Type of acquisiton; either over COM, or from data files, or a fake acquisition for testing."
Tooltip_DumpAcquiredData = "When this checkbox is enabled, in addition to writing HDF5 files, a copy of the strings received from the box will be saved in the directory " + "\"" + stringDumpFolderName + "\" under your data directory."
Tooltip_ParanoidMode     = "Paranoid mode initiates a less performant data acquisition model but safer. It reads lines from the device line by line.\n" \
                           "In contrast, non-paranoid mode empties the buffer of the device and parses it. Use this if you think the software has a problem\n" \
                           "with standard acquisition mode. This mode works only for COM acquisition."
Tooltip_StartAcquisition = "Start data acquisition from the device based on the configuration you chose. "
Tooltip_MissingBatches   = "Due to data communication issues, data may be lost within communication. This is an estimate of the number of batches missing during communication."

if hasattr(sys, 'frozen'):
    ScriptPath = os.path.join(os.path.dirname(sys.executable),"GNOMEAcquirerLibs")
else:
    ScriptPath = os.path.dirname(os.path.realpath(__file__))


class AcqusitionWindow(QWidget):
    #The following signal definition is per object, nevertheless it has to be here.
    #Qt takes a copy of it for each object
    stopAcquisitionSignal = QtCore.Signal()
    # BatchesToStoredInSingleFile = 60
    BatchSecondsLength = 1

    dlProgress_update = pyqtSignal(float)
    dlProgress_done = pyqtSignal()


    def __init__(self, parent = None):
        super(AcqusitionWindow, self).__init__(parent)

        self.splash_pix = QPixmap(os.path.join(ScriptPath, 'icon.png'))
        self.splash = QSplashScreen(self.splash_pix, Qt.WindowStaysOnTopHint)
        self.splash.setMask(self.splash_pix.mask())
        self.splash.show()
        QCoreApplication.processEvents()

        is_windows = sys.platform.startswith('win')

        self.logFilePath = ""
        self.viewLogWidget = ViewLog()

        self.dataCollected = {}

        #set window icon
        self.windowIcon = QIcon(os.path.join(ScriptPath, "icon.png"))
        self.setWindowIcon(self.windowIcon)
        self.setWindowTitle("GNOME Data Acquisition Software")

        #initiate tabs widget
        self.datasetsTabs   = QTabWidget()
        self.listOfDatasets = {}

        #add/remove datasets button objects
        self.addDatasetButton    = QPushButton("Add dataset")
        self.addDatasetButton.setToolTip(Tooltip_AddDataset)
        self.removeDatasetButton = QPushButton("Remove dataset")
        self.removeDatasetButton.setToolTip(Tooltip_RemoveDataset)

        #Splitter of the two halves of the main window
        self.mainWidgetsSplitter = QSplitter()
        self.mainWidgetsSplitter.setOrientation(Qt.Horizontal)
        self.mainLeftWidget = QWidget()
        self.mainLeftGridLayout = QGridLayout()
        self.mainRightWidget = QWidget()
        self.mainRightGridLayout = QGridLayout()
        self.mainWidgetsSplitter.addWidget(self.mainLeftWidget)
        self.mainWidgetsSplitter.addWidget(self.mainRightWidget)
        self.mainLeftWidget.setLayout(self.mainLeftGridLayout)
        self.mainRightWidget.setLayout(self.mainRightGridLayout)

        #data path groupbox
        self.dataGroupbox = QGroupBox()
        self.dataGroupbox.setTitle("Data path")
        self.dataGroupbox.setToolTip(Tooltip_DataPath)
        self.dataLayout = QGridLayout()
        self.dataGroupbox.setLayout(self.dataLayout)

        #Acquisition type (acquire, read files, fake)
        self.acqTypeOptions = \
            {
                0: AcquisitionTypeOption_COM,
                1: AcquisitionTypeOption_Fake,
                2: AcquisitionTypeOption_File
            }
        self.acquisitionFromFilesPath = ""

        self.inputOptionsGroupbox       = QGroupBox("Input options")
        self.inputOptionsGroupboxLayout = QGridLayout()
        self.stationNameLine       = QLineEdit()
        self.stationNameLabel      = QLabel("Station name: ")
        self.inputOptionsGroupbox.setLayout(self.inputOptionsGroupboxLayout)
        self.inputOptionsGroupboxLayout.addWidget(self.stationNameLabel,0,0,1,1)
        self.inputOptionsGroupboxLayout.addWidget(self.stationNameLine,0,1,1,3)

        self.stationNameTooltipText = Tooltip_StationName
        self.stationNameLine.setToolTip(self.stationNameTooltipText)
        self.stationNameLabel.setToolTip(self.stationNameTooltipText)

        self.portLabel = QLabel("Device:")
        self.portLine  = QComboBox()
        self.portLabel.setToolTip(Tooltip_DeviceName)
        self.portLine.setToolTip(Tooltip_DeviceName)
        self.refreshDevicesButton = QPushButton("Refresh devices")

        self.acquisitionTypeLabel = QLabel("Acquisition type: ")
        self.acquisitionTypeLine  = QComboBox()
        [self.acquisitionTypeLine.addItem(self.acqTypeOptions[i]) for i in self.acqTypeOptions]
        self.acquisitionTypeLabel.setToolTip(Tooltip_AcquisitionType)
        self.acquisitionTypeLine.setToolTip(Tooltip_AcquisitionType)
        self.inputOptionsGroupboxLayout.addWidget(self.acquisitionTypeLabel,1,0,1,1)
        self.inputOptionsGroupboxLayout.addWidget(self.acquisitionTypeLine ,1,1,1,3)
        self.inputOptionsGroupboxLayout.addWidget(self.portLabel,2,0,1,1)
        self.inputOptionsGroupboxLayout.addWidget(self.portLine ,2,1,1,2)
        self.inputOptionsGroupboxLayout.addWidget(self.refreshDevicesButton ,2,3,1,1)


        self.dataPathLine         = QLineEdit()
        self.dataPathBrowseButton = QPushButton("Browse")
        self.dataPathLine.setToolTip(Tooltip_DataPath)
        self.dataPathBrowseButton.setToolTip(Tooltip_DataPath)
        self.dataPathBrowseButton.clicked.connect(self.browseDataFolder)
        self.dataLayout.addWidget(self.dataPathLine,0,0,1,1)
        self.dataLayout.addWidget(self.dataPathBrowseButton,0,1,1,1)

        self.datasetsGroupbox = QGroupBox()
        self.datasetsGroupbox.setTitle("Datasets")
        self.datasetsGroupboxLayout = QGridLayout()
        self.datasetsGroupboxLayout.addWidget(self.addDatasetButton,0,0,1,1)
        self.datasetsGroupboxLayout.addWidget(self.removeDatasetButton,0,1,1,1)
        self.datasetsGroupboxLayout.addWidget(self.datasetsTabs,1,0,1,2)
        self.datasetsGroupbox.setLayout(self.datasetsGroupboxLayout)

        self.startAcqText = "Start data acquisition"
        self.stopAcqText  = "Stop data acquisition"
        self.startStopButtonWaitText = "Please wait..."
        self.acquisitionRunning = False

        self.startAcquisitionButton = QPushButton(self.startAcqText)
        self.startAcquisitionButton.setToolTip(Tooltip_StartAcquisition)


        #dump acquired strings
        self.dumpAcquiredStringsToFileCheckbox = QCheckBox("Dump acquired strings to the folder \"" + stringDumpFolderName + "\" (watch for diskspace)")
        self.dumpAcquiredStringsToFileCheckbox.setToolTip(Tooltip_DumpAcquiredData)

        #paranoid mode checkbox
        self.paranoidModeCheckbox = QCheckBox("Paranoid acquisition mode (less efficient)")
        self.paranoidModeCheckbox.setToolTip(Tooltip_ParanoidMode)

        #data analysis station
        self.dataAnalysisGroupBox = QGroupBox()
        self.dataAnalysisGroupBox.setTitle("Data view")
        self.dataAnalysisLayout   = QGridLayout()
        self.dataAnalysisGroupBox.setLayout(self.dataAnalysisLayout)


        self.dataAnalysisWidget = DataAnalysisWidget()

        # add the plot canvas to a window
        self.dataAnalysisLayout.addWidget(self.dataAnalysisWidget)

        self.aboutButton = QPushButton("About")
        self.checkForUpdatesButton = QPushButton("Check for updates")

        #Logging groupbox
        self.loggingGroupBox = QGroupBox()
        self.loggingGroupBox.setTitle("Acquired data status")
        self.loggingLayout   = QGridLayout()
        self.loggingGroupBox.setLayout(self.loggingLayout)

        self.missingBatchesLabel = QLabel("Missing batches: ")
        self.missingBatchesLine  = QLineEdit()
        self.missingBatchesResetButton = QPushButton("Reset counter")
        self.viewLogFileButton = QPushButton("View log file")
        self.missingBatchesLine.setDisabled(0)
        self.missingBatchesLabel.setToolTip(Tooltip_MissingBatches)
        self.missingBatchesLine.setToolTip(Tooltip_MissingBatches)
        self.missingBatchesLine.setReadOnly(1)
        self.missingBatchesResetButton.setToolTip(Tooltip_MissingBatches)
        self.loggingLayout.addWidget(self.missingBatchesLabel, 0, 0, 1, 1)
        self.loggingLayout.addWidget(self.missingBatchesLine, 0, 1, 1, 1)
        self.loggingLayout.addWidget(self.missingBatchesResetButton, 0, 2, 1, 1)
        self.loggingLayout.addWidget(self.viewLogFileButton, 0, 3, 1, 1)
        self.resetMissingBatches()
        self.missingBatchesResetButton.clicked.connect(self.resetMissingBatches)
        self.viewLogFileButton.clicked.connect(self.viewLogFile)


        self.mainLayout = QGridLayout()
        self.mainLayout.addWidget(self.mainWidgetsSplitter)

        self.mainLeftGridLayout.addWidget(self.dataGroupbox,0,0,1,4)
        self.mainLeftGridLayout.addWidget(self.inputOptionsGroupbox,1,0,1,4)
        self.mainLeftGridLayout.addWidget(self.datasetsGroupbox,3,0,1,4)
        self.mainLeftGridLayout.addWidget(self.dumpAcquiredStringsToFileCheckbox,4,0,1,4)
        self.mainLeftGridLayout.addWidget(self.paranoidModeCheckbox,5,0,1,4)
        self.mainLeftGridLayout.addWidget(self.startAcquisitionButton,6,0,1,4)
        if is_windows:
            # if hasattr(sys, 'frozen'):
            self.mainLeftGridLayout.addWidget(self.checkForUpdatesButton, 7, 0, 1, 2)
            self.mainLeftGridLayout.addWidget(self.aboutButton, 7, 2, 1, 2)
            # else:
            #     self.mainLeftGridLayout.addWidget(self.aboutButton, 7, 0, 1, 4)
        else:
            self.mainLeftGridLayout.addWidget(self.aboutButton, 7, 0, 1, 4)

        self.mainRightGridLayout.addWidget(self.dataAnalysisGroupBox,0,4,6,1)
        self.mainRightGridLayout.addWidget(self.loggingGroupBox,6,4,2,1)


        self.addDatasetButton.clicked.connect(self.addDataset)
        self.removeDatasetButton.clicked.connect(self.removeDataset)
        self.startAcquisitionButton.clicked.connect(self.startAcquisitionButtonAction)
        self.acquisitionTypeLine.currentIndexChanged.connect(self.acquisitionTypeChanged)
        self.aboutButton.clicked.connect(self.showAboutDialog)
        self.checkForUpdatesButton.clicked.connect(self.updateSoftwareSlot)

        #refresh devices list and create button response
        self.refreshDevices()
        self.refreshDevicesButton.clicked.connect(self.refreshDevices)

        self.setLayout(self.mainLayout)

        self.screen = QDesktopWidget().screenGeometry()

        self.resize(int(round(self.screen.width()*0.75)),int(round(self.screen.height()*0.75)))

        self.addDatasetWithName(defaultDatasetName,True,True)
        self.addDatasetWithName(sanityDatasetName,True,True)

        #initialize thread object, in order to check whether it exists on program exit
        #see the member function closeEvent()
        self.acquirer = None

        #plot lock as mutex
        self.plotLocked = False

        #skip plots to avoid having the window freeze due to rapid plotting
        #The next field is: How many batches to process before updating the plot
        #a very small number like "1" would make the refresh rate of the plot too high and the program will become slow
        # and unable to catch up with the data pulled from the device
        self.skipPlotEvery = 5
        #initial counter value
        self.skipPlotCounter = 0 #do not change this

        # self.dataPathLine.setText("D:\gnome\clientTests")
        # self.stationNameLine.setText("mainz01")

        self.loadSettingsFromFile(ConfigFileName)

        self.lastAttrs = {}

        try:
            cfgtestfile = "tmpcfgtest.bin"
            self.saveSettingsToFile(cfgtestfile)
            os.remove(cfgtestfile)
        except:
            QMessageBox.information(self,"Error","Unable to save configuration file in the program path: " +
            os.path.dirname(os.path.realpath(__file__)) + "\n\n" +
            "This means that you have to redo the configuration of your acquisition every time you start the program.\n\n"
            "To resolve this, make sure you have write access to that path. It is possible that you have the program installed on drive C:, "
            "which is considered a secured system drive in Windows versions newer than Vista. Try running the program as admin "
            "by right clicking on the program and selecting \"Run as administrator\"")

        if is_windows:
            self.updateTimer = QTimer(self)

            self.updateTimer.timeout.connect(self.updateSoftwareSlot_silent)
            self.updateTimer.start(15*60*1000) #check for updates once per hour

            self.updateSoftware(True)


    def doAfterShown(self):
        self.splash.finish(self)
        QCoreApplication.processEvents()
        QMessageBox.information(self, "Information",
                                "Please be aware that the program will not stop the acquisition when errors occur. "
                                "The program will merely continuously skip the data that has the problem. If you notice that all data is being skipped (a counter is provided for that purpose), then review the log and check the cause of the problem. "
                                "This is due to issues related to data communication reliability. Until these issues are resolved, there is no way to tell the difference between corrupt data and user input problems in the acquisition box.")



    def updateSoftwareSlot(self):
        self.updateSoftware(False)

    def updateSoftwareSlot_silent(self):
        self.updateSoftware(True)

    def updateSoftware(self, silent):
        # if not hasattr(sys, 'frozen'):
        #     return

        try:
            updatesReturn = CheckForUpdates()
        except Exception as e:
            if not silent:
                QMessageBox.information(self, "Update error", "Error while checking for updates. Are you connected to the internet?\n\nException says: " + str(e))

            traceback.print_exc()
            print("Error while checking for updates. Are you connected to the internet? Exception says: " + str(e))
            return

        if updatesReturn["success"] is False:
            if not silent:
                QMessageBox.information(self, "Update error", "Checking for updates failed. Exception says: " +
                                        str(updatesReturn["msg"]))
            print("Checking for updates failed. Exception says: " + str(updatesReturn["msg"]))
            return

        if updatesReturn['newerAvailable'] is False:
            if not silent:
                QMessageBox.information(self, "Update info", "You have the latest version. No need for updates.")

        else:
            self.checkForUpdatesButton.setStyleSheet('QPushButton {color: red;}')
            self.checkForUpdatesButton.setText("Update is available!")
            self.updateTimer.stop()
            if not silent:
                is_windows = sys.platform.startswith('win')
                is_frozen = hasattr(sys, 'frozen')

                if is_windows and is_frozen:
                    reply = QMessageBox.question(self, 'Update available',
                                             "An update to version " + updatesReturn['version'] + " is available. Would you like to download it now?",
                                             QMessageBox.Yes, QMessageBox.No)
                    if reply == QMessageBox.No:
                        return

                    self.doUpdateSoftware(updatesReturn)
                else:
                    QMessageBox.information(self, "Update info", "There is an update available. "
                                                                 "Please upgrade this software ASAP using your "
                                                                 "Python package manager.\n"
                                                                 "For Anaconda, run:\n"
                                                                 " conda upgrade gnomeacquirer\n"
                                                                 "For pip, run:\n"
                                                                 " pip3 install gnomeacquirer -U")

    def doUpdateSoftware(self,updatesReturn):
        updateFilePath = QFileDialog.getSaveFileName(self, "Save downloaded file...",
                                                     os.path.join(os.path.expanduser("~\\Downloads\\"),
                                                                  updatesReturn['url'].split('/')[-1]),
                                                     "All Files (*)")
        # sometimes QFileDialog.getSaveFileName returns a tuple with the filter as the second element
        if type(updateFilePath) == tuple:
            updateFilePath = updateFilePath[0]

        if updateFilePath != "":
            downloadDialogStaticText = "Please wait while GNOME Acquirer version " + updatesReturn['version'] + " is downloaded...\n"
            downloadDialog = QProgressDialog(self)
            downloadDialog.setWindowTitle("Downloading update...")
            screen = QApplication.desktop().screenNumber(QApplication.desktop().cursor().pos())
            screenRes = QDesktopWidget().screenGeometry(screen)

            downloadDialog.resize(screenRes.width() / 2, downloadDialog.height())
            downloadDialog.setModal(True)
            downloadDialog.setLabelText(downloadDialogStaticText)
            centerPoint = QDesktopWidget().screenGeometry(screen).center()
            centerPoint.setX(centerPoint.x() - downloadDialog.width() / 2)
            centerPoint.setY(centerPoint.y() - downloadDialog.height() / 2)
            downloadDialog.move(centerPoint)

            downloadDialog.show()

            downloadDialog.isDownloadCanceled = False
            downloadDialog.isFinishActionDone = False

            def setDownloadProgressValue(val):
                downloadDialog.setValue(val)
                downloadDialog.setLabelText(downloadDialogStaticText +
                                            str(human_readable_bytes(downloadDialog.downloadedFileSize)) + " / " +
                                            str(human_readable_bytes(downloadDialog.totalFileSize)))

            def cancelDownload():
                downloadDialog.isDownloadCanceled = True

            def downloadDone():
                print("Update download finished.")

                reply = QMessageBox.question(self, 'Download finished!',
                                             "Update downloaded. Would you like to run it? (the current session will be closed)",
                                             QMessageBox.Yes, QMessageBox.No)
                if reply == QMessageBox.Yes:
                    try:
                        os.startfile(updateFilePath) #run the downloaded file
                        sys.exit(0)
                    except Exception as e:
                        QMessageBox.information(self, "Cannot start installer",
                                                "Unable to start installer. Reason: " + str(e))

            self.dlProgress_update.connect(setDownloadProgressValue)
            downloadDialog.canceled.connect(cancelDownload)
            self.dlProgress_done.connect(downloadDone)

            def report(block_count, block_size, total_size):
                QCoreApplication.processEvents()
                downloadDialog.totalFileSize = total_size
                downloadDialog.downloadedFileSize = block_count * block_size
                if downloadDialog.isDownloadCanceled:
                    raise Exception("User canceled the download.")
                else:
                    progVal = float((100 * block_count * block_size) / total_size)
                    if block_count == 0:
                        # print("block_count == 0")
                        self.dlProgress_update.emit(0)
                    # if (block_count * block_size) >= total_size:
                    #     self.dlProgress_done.emit()
                    #     downloadDialog.isFinishActionDone = True
                    # print("BS={0} TS={1} incAmount={2}".format(block_size,total_size,incAmount))
                    else:
                        self.dlProgress_update.emit(progVal)

            try:
                socket.setdefaulttimeout(10)
                urllib.request.urlretrieve(updatesReturn['url'], updateFilePath, reporthook=report)
            except Exception as e:
                print("Download stopped. Reason: " + str(e))
                downloadDialog.isDownloadCanceled = True
                QMessageBox.information(self, "Download stopped", "Download stopped. Reason: " + str(e))

            if not downloadDialog.isDownloadCanceled:
                self.dlProgress_done.emit()

            #cleanup
            urllib.request.urlcleanup()
            #since signals are bound to the main window, they have to be disconnected manually
            self.dlProgress_done.disconnect(downloadDone)
            self.dlProgress_update.disconnect(setDownloadProgressValue)
            downloadDialog.destroy()
            print(updateFilePath)
        else:
            print("No path given")

    def viewLogFile(self):
        if(self.viewLogWidget.setLogFile(self.logFilePath) >= 0):
            self.viewLogWidget.show()

    def showAboutDialog(self):
        QMessageBox.about(self,"About GNOME Data Acquirer","GNOME Data Acquirer, version " + SOFTWARE_VERSION + "\n\n"
            "This program was designed by the Budker Group in Mainz, by Samer Afach. \n"
            "In case you find bugs or you have suggestions, send an e-mail to afach@uni-mainz.de")

    def isFolderWritable(self,path):
        try:
            testfile = tempfile.TemporaryFile(dir = path)
            testfile.close()
        except:
            return False
        return True

    def isDataPathValid(self):
        path = self.dataPathLine.text()
        if path != "":
            if self.isFolderWritable(path):
                return True
            else:
                return False
        else:
            return False

    def browseDataFolder(self):
        self.dataPathLine.setText(QFileDialog.getExistingDirectory(self, "Open Directory", "C:", QFileDialog.ShowDirsOnly))

    def resetMissingBatches(self):
        self.missingBatches = 0
        self.updateMissingBatchesField()

    def updateMissingBatchesField(self):
        self.missingBatchesLine.setText(str(self.missingBatches))

    def addMissingBatches(self,missingBatches):
        self.missingBatches += missingBatches
        self.updateMissingBatchesField()

    def addDataset(self):
        datasetName = QInputDialog.getText(self,"Dataset name?","Enter dataset name (CamelCase, no spaces)")[0].replace(" ","").replace("\t","")
        if not datasetName == "":
            if re.match(dataset_regex, datasetName) is None:
                QMessageBox.information(self, "Error",
                                        "Dataset name \"" + datasetName + "\""
                                        "is invalid. It should start with letters, and can contain numbers"
                                        " and cannot have more than one underscore consecutively.")
                return

            #forbid forbidden words
            for word in ForbiddenDatasetNames:
                if len(datasetName) >= len(word):
                    if datasetName[:len(word)] == word:
                        QMessageBox.information(self,"Error","Dataset name cannot be \"" + word + "\". This word is reserved.")
                        return

            for i in range(self.datasetsTabs.count()):
                self.tmpTabName = self.datasetsTabs.tabText(i)
                if datasetName == self.tmpTabName:
                    QMessageBox.information(self,"Error","A dataset with name " + datasetName + " already exists. "
                                                         "Please choose a different name.")
                    return
            self.addDatasetWithName(datasetName,False,False)

        if self.acqTypeOptions[self.acquisitionTypeLine.currentIndex()] == AcquisitionTypeOption_Fake:
            self.acquisitionTypeChanged(self.acquisitionTypeLine.currentIndex())

    def addDatasetWithName(self,datasetName,isLocked,isvaluelocked):
        for i in range(self.datasetsTabs.count()):
            tabName = self.datasetsTabs.tabText(i)
            if datasetName == tabName:
                #if dataset already exists, just return
                return False
        self.listOfDatasets[datasetName] = DatasetForm(datasetName, parent=self)
        self.listOfDatasets[datasetName].isLocked      = isLocked
        self.listOfDatasets[datasetName].isValueLocked = isvaluelocked
        self.datasetsTabs.addTab(self.listOfDatasets[datasetName],datasetName)
        return True

    def refreshDevices(self):
        self.devicesNames = []
        try:
            #remove all currently listed device
            while self.portLine.count() > 0:
                self.portLine.removeItem(0)

            self.devicesNames = get_serial_ports()
            # if no seerial devices found
            if len(self.devicesNames) == 0:
                self.portLine.addItem(NoDevicesString)
            else:
                self.portLine.addItems(self.devicesNames)

            self.dealWithNoDevicesFound()

        except Exception as e:
            self.portLine.setDisabled(1)
            self.portLine.addItem(NoDevicesString)
            QMessageBox.information(self,"Error","Error while trying to find serial devices. Make sure you have devices connected. Exception message: " + str(e))

    def dealWithNoDevicesFound(self):
        '''
        Decides whether the devices combobox has to be disabled
        :return: None
        '''
        if len(self.devicesNames) == 0:
            self.portLine.setDisabled(1)
        else:
            self.portLine.setDisabled(0)

    def removeDataset(self):
        if self.datasetsTabs.count() > 0:
            items = []
            itemsDic = {}
            for i in range(self.datasetsTabs.count()):
                tabName = self.datasetsTabs.tabText(i)
                items.append(tabName)
                itemsDic[tabName] = i
            self.datasetReturn = QInputDialog.getItem(self,"Dataset name?","Choose dataset to remove",items,self.datasetsTabs.currentIndex(),False)
            self.datasetName   = self.datasetReturn[0]
            self.datasetChosen = self.datasetReturn[1]
            if self.datasetChosen:
                if not self.datasetsTabs.widget(itemsDic[self.datasetName]).isLocked:
                    self.datasetsTabs.removeTab(itemsDic[self.datasetName])
                else:
                    QMessageBox.information(self,"Error","This dataset is locked. You cannot remove it, as data will then not comply to the standard of GNOME.")
        else:
            QMessageBox.information(self,"Error","You can't ask to remove datasets when no datasets are present")


    def closeEvent(self, event):
        while self.datasetsTabs.count():
            self.datasetsTabs.removeTab(0)

        while self.mainLayout.count():
            item = self.mainLayout.takeAt(0)
            widget = item.widget()
            if widget:
                try:
                    widget.close()
                except:
                    pass

    def healthyWait(self,totalWait):
        self._waitTime = 0
        self._waitStep = 0.1
        if totalWait < self._waitStep:
            time.sleep(totalWait)
        else:
            while self._waitTime < totalWait:
                time.sleep(self._waitStep)
                self._waitTime += self._waitStep
                QCoreApplication.processEvents()

    def startAcquisitionButtonAction(self):
        for i in range(self.datasetsTabs.count()):
            tabName = self.datasetsTabs.tabText(i)
            if re.match(dataset_regex, tabName) is None:
                QMessageBox.information(self, "Error",
                                        "Dataset name \"" + tabName +  "\" is invalid. It should start with letters, "
                                        "and can contain numbers"
                                        " and cannot have more than one underscore consecutively.")
                return

        QCoreApplication.processEvents()
        if self.acquisitionRunning:
            #lock acquisition start button
            self.startAcquisitionButton.setDisabled(1)
            QCoreApplication.processEvents()
            self.doStopAcquisition()
            QCoreApplication.processEvents()

            #unlock acquisition start button
            #the unlock will be done by the thread

        else:
            if self.isDataPathValid():
                self.startAcquisitionButton.setDisabled(1)
                QCoreApplication.processEvents()
                self.doStartAcquisition()
                QCoreApplication.processEvents()
                QCoreApplication.processEvents()
            else:
                QMessageBox.information(self,"Error","Error: The data path is either invalid or non-writable. Please select a directory you selected is valid and where you can write.")

    def doStopAcquisition(self):
        self.acquisitionRunning = False
        self.startAcquisitionButton.setText(self.startAcqText)
        self.addDatasetButton.setDisabled(0)
        self.portLine.setDisabled(0)
        self.refreshDevicesButton.setDisabled(0)
        self.removeDatasetButton.setDisabled(0)
        for i in range(self.datasetsTabs.count()):
            # self.datasetsTabs.widget(i).setDisabled(0)
            self.datasetsTabs.widget(i).disableControls(False)
        self.acquisitionTypeLine.setDisabled(0)
        self.dumpAcquiredStringsToFileCheckbox.setDisabled(0)
        self.paranoidModeCheckbox.setDisabled(0)
        self.stationNameLine.setDisabled(0)
        self.dataPathLine.setDisabled(0)
        self.stopAcquisitionSignal.emit()
        self.dataPathBrowseButton.setDisabled(0)
        self.dealWithNoDevicesFound()

    def doStartAcquisition(self):
        self.acquisitionRunning = True
        self.startAcquisitionButton.setText(self.stopAcqText)
        self.addDatasetButton.setDisabled(1)
        self.portLine.setDisabled(1)
        self.refreshDevicesButton.setDisabled(1)
        self.removeDatasetButton.setDisabled(1)
        for i in range(self.datasetsTabs.count()):
            # self.datasetsTabs.widget(i).setDisabled(1)
            self.datasetsTabs.widget(i).disableControls(True)
        self.acquisitionTypeLine.setDisabled(1)
        self.dumpAcquiredStringsToFileCheckbox.setDisabled(1)
        self.paranoidModeCheckbox.setDisabled(1)
        self.stationNameLine.setDisabled(1)
        self.dataPathLine.setDisabled(1)
        self.dataPathBrowseButton.setDisabled(1)

        self.acquisitionCall()

    def getAttributesFromGUI(self):
        attrsContainer = {}
        for i in range(self.datasetsTabs.count()):
            ds_name = self.datasetsTabs.tabText(i)
            num_rows = self.datasetsTabs.widget(i).attribsTable.rowCount()
            for j in range(num_rows):
                if not (self.datasetsTabs.tabText(i) in attrsContainer):
                    attrsContainer[self.datasetsTabs.tabText(i)] = {}

                name  = self.datasetsTabs.widget(i).attribsTable.item(j, self.datasetsTabs.widget(i).AttributesColumns[Attribute_NameKey]).text()
                type  = self.datasetsTabs.widget(i).attribsTable.item(j, self.datasetsTabs.widget(i).AttributesColumns[Attribute_TypeKey]).text()
                value = self.datasetsTabs.widget(i).attribsTable.item(j, self.datasetsTabs.widget(i).AttributesColumns[Attribute_ValueKey]).text()
                attrsContainer[self.datasetsTabs.tabText(i)].update(
                    {
                        name: {Attribute_NameKey:  name,
                               Attribute_ValueKey: value,
                               Attribute_TypeKey:  type}
                    })

            # this older version takes values not from the GUI, but from the table stored, attributesList

            # for attName in sorted(self.datasetsTabs.widget(i).attributesList):
            #     # if dict doesn't exist for this dataset, create it!
            #     if not (self.datasetsTabs.tabText(i) in attrsContainer):
            #         attrsContainer[self.datasetsTabs.tabText(i)] = {}
            #
            #     # append attribute to dict of attributes
            #     attrsContainer[self.datasetsTabs.tabText(i)].update(
            #         {
            #             attName: {Attribute_NameKey:  attName,
            #                       Attribute_ValueKey: self.datasetsTabs.widget(i).attributesList[attName].Value,
            #                       Attribute_TypeKey:  self.datasetsTabs.widget(i).attributesList[attName].Type}
            #         })

        return attrsContainer


    def updateAttributesDuringAcquisition(self):
        """
        This function updates the attributes from the DatasetForms to a container that is passed to the object
        that writes the files
        :return:
        """
        self.dataCollected.acqConfig.attributes = self.getAttributesFromGUI()

    def acquisitionCall(self):
        self.acqConfig = AcquisitionConfig()
        self.acqConfig.registerDevice(self)
        self.acqConfig.dataPath = self.dataPathLine.text()
        self.acqConfig.stationName = self.stationNameLine.text()
        self.acqConfig.logFilePath = os.path.join(self.acqConfig.dataPath, logDirName,
                                        logFileNamePrefix +
                                        datetime.datetime.utcnow().strftime(logDateTimeFormat) +
                                        logFileNameSuffix)
        self.logFilePath = self.acqConfig.logFilePath
        self.acqConfig.attributes = self.getAttributesFromGUI()

        #initialize local data queues, for the data that will be received from the device
        self.dataCollected = DataCollected(self.acqConfig)

        self.acqConfig.registerChannels(self)
        if self.acqConfig.errorFlag:
            QMessageBox.information(self,"Error",self.acqConfig.errorMessage)
            self.doStopAcquisition()
            self.reenableStartButton()
            return

        self.acqConfig.isFakeAcquisition   = (self.acqTypeOptions[self.acquisitionTypeLine.currentIndex()] == AcquisitionTypeOption_Fake)
        self.acqConfig.isFileAcquisition   = (self.acqTypeOptions[self.acquisitionTypeLine.currentIndex()] == AcquisitionTypeOption_File)
        self.acqConfig.doDumpStringsToFile = self.dumpAcquiredStringsToFileCheckbox.isChecked()
        self.acqConfig.paranoidMode        = self.paranoidModeCheckbox.isChecked()
        self.acqConfig.fileAcqPath         = self.acquisitionFromFilesPath
        self.dataAnalysisWidget.initPlots(self.datasetsTabs.count()+1)

        if (not self.acqConfig.isFileAcquisition) and (not self.acqConfig.isFakeAcquisition):
            if self.portLine.currentText() == NoDevicesString:
                QMessageBox.information(self, "Error",
                                        "A valid device is not selected. "
                                        "Please select a valid device before starting acquisition")
                self.doStopAcquisition()
                self.reenableStartButton()
                return

        self.acquirer = DataAcquirer(self.acqConfig)
        self.stopAcquisitionSignal.connect(self.acquirer.stopAcquisition)
        self.acquirer.reenableStartButtonSignal.connect(self.reenableStartButton)
        self.acquirer.finishedSignal.connect(self.reenableStartButton)
        self.acquirer.postErrorMessage.connect(self.acquisitionThreadError)
        self.acquirer.postResultsSignal.connect(self.processDataFromThreadSlot)
        self.acquirer.postNumMissingBatches.connect(self.addMissingBatches)
        castingReturn = self.testAttributesCasting()

        if castingReturn:
            self.saveSettingsToFile(ConfigFileName)
            # self.acquirer.start()
            self.acquirer.run()
        else:
            self.doStopAcquisition()
            self.reenableStartButton()

        return

    def getMagneticFieldEquationString(self):
        attribs = self.getAttributesFromGUI()
        return attribs[defaultDatasetName][defaultMagFieldEqAttrName][Attribute_ValueKey]

    def getDatasetsList(self):
        listOfDS = []
        keysList = self.acqConfig.requestedChannelsTypes.keys()
        for key in keysList:
            listOfDS.append(key)
        return listOfDS

    def testAttributesCasting(self):
        print("Testing attribute casting...")
        attribs = self.getAttributesFromGUI()
        print("Attributes received: ", attribs)
        tmpPath = copy.deepcopy(self.dataPathLine.text())
        tmpPath = os.path.join(tmpPath, "test.h5")
        self.h5fObj = h5py.File(tmpPath, "w")
        for i in range(self.datasetsTabs.count()):
            ds_name = self.datasetsTabs.tabText(i)
            self.tmpDataset = self.h5fObj.create_dataset(ds_name, data=[np.int32(1)])
            print("Dataset name: " + str(self.tmpDataset.name))
            for attName in sorted(attribs[ds_name]):
                if re.match(attrib_regex, attName) is None:
                    QMessageBox.information(self, "Error",
                                            "Attribute name \"" + attName + "\" in dataset " + ds_name +
                                            " is invalid. It should start with letters, and can contain numbers"
                                            " and cannot have more than one underscore consecutively.")
                    self.h5fObj.close()
                    return False

                try:
                    self.tmpDataset.attrs[attName] = \
                    ApplyAttributeType(ApplyAttributeReplacement_tester(attribs[ds_name][attName]["Value"]),
                                       attribs[ds_name][attName]["Type"])
                except Exception as e:
                    QMessageBox.information(self, "Error", "Error while trying to cast the parameter " + attName +
                                            " in dataset " + ds_name + " to type " +
                                            attribs[ds_name][attName]["Type"] + ". "
                                            "Make sure that the type you chose for this variable is compatible "
                                                                    "with the its contents. The contents "
                    "are: " + ApplyAttributeReplacement_tester(str(attribs[ds_name][attName]["Value"])) + "\n"
                    "Exception says: " + str(e))
                    self.h5fObj.close()
                    try:
                        os.remove(tmpPath)
                    except: pass
                    return False
        print("Attributes casting tests successful!")
        self.h5fObj.close()
        try:
            os.remove(tmpPath)
        except: pass

        # parse magnetic field equation and test it

        def attr_getter(attname: str) -> str:
            # in case of failure, this value stores the last known value
            parts = attname.split("/")
            if len(parts) == 1:
                raise SyntaxError("Error: Magnetic field equation given with a global attribute. "
                                  "This is not supported in this software!")

            elif len(parts) == 2:
                for i in range(self.datasetsTabs.count()):
                    print(self.datasetsTabs.tabText(i), parts[0])
                    if self.datasetsTabs.tabText(i) == parts[0]:
                        val = attribs[self.datasetsTabs.tabText(i)][parts[1]]["Value"]
                        val = ApplyAttributeReplacement_tester(val)
                        return val

            else:
                raise SyntaxError("Error: Attributes should be given in the form: ${dataset/attr}")

        try:
            magFieldEquationStr = self.getMagneticFieldEquationString()

            eq_info = gd.parse_equation(magFieldEquationStr)
            print(eq_info)
            if not eq_info["success"]:
                raise SyntaxError(eq_info["msg"])
            magFieldEquationStr = eq_info["equation"]

            datasetsVarReps = {}
            attrsVarReps = {}
            listOfDatasets = self.getDatasetsList()
            if eq_info["datasets_indices"] != eq_info["datasets"]:
                raise SyntaxError("All datasets in this software are one dimensional. "
                                  "You can't use deindexing operators, such as dataset[[i]].")

            for ds in eq_info["datasets"]:
                if ds not in listOfDatasets:
                    raise SyntaxError("Your equation contains the dataset " + ds + ", which doesn't seem to be within "
                                      "the datasets you have in this acquisition.")

            for key in listOfDatasets:
                datasetsVarReps[key] = float(np.random.rand())
            for attr in eq_info["attrs"]:
                val = attr_getter(attr)
                if val is None:
                    raise SyntaxError("Unable to substitute attribute " + attr + ". Are you sure it exists?")
                else:
                    attrsVarReps[attr] = val

            # recursive replacement of attributes
            prev_eq = magFieldEquationStr
            while True:
                for attr in attrsVarReps:
                    magFieldEquationStr = magFieldEquationStr.replace("${"+attr+"}", attrsVarReps[attr])

                if magFieldEquationStr == prev_eq:
                    break
                else:
                    prev_eq = magFieldEquationStr
            expression  = sympy.S(magFieldEquationStr)
            all_symbols = [str(x) for x in expression.atoms(sympy.Symbol)]
            if not set(all_symbols).issubset(listOfDatasets):
                QMessageBox.information(self, "Magetic field equation variables error",
                                              "The magnetic field equation provided: \"" + magFieldEquationStr + "\" must have only dataset names as variables.\n"
                                              "This is because the magnetic field is supposed to be generated from data points from recorded datasets.")
                return False

            np.double(expression.subs(datasetsVarReps))

        except Exception as e:
            QMessageBox.information(self,"Magnetic field equation parsing error",
                                        "Error while parsing the magnetic field equation provided. \n"
                                        "Please make sure the equation provided complies to sympy standards. Exception says: " + str(e))
            print("Magnetic field equation parsing error","Error while parsing the magnetic field equation provided. \n"
                                        "Please make sure the equation provided complies to sympy standards. Exception says: " + str(e))
            return False

        return True

    #saves settings to config file
    def saveSettingsToFile(self, filename):
        self.configFileObj = None
        try:
            print("Saving config file: " + filename)
            self.configFileObj = h5py.File(filename,"w")
            self.configFileObj.attrs[ConfigOptionName_DataPath]    = self.dataPathLine.text()
            self.configFileObj.attrs[ConfigOptionName_StationName] = self.stationNameLine.text()
            self.configFileObj.attrs[ConfigOptionName_PortName] = self.portLine.currentText()

            for i in range(self.datasetsTabs.count()):
                self.tmpDataset = self.configFileObj.create_dataset(self.datasetsTabs.tabText(i),data=[np.int32(1)])
                counter = 0
                for attName in sorted(self.datasetsTabs.widget(i).attributesList):
                    #only attributes that are not locked are saved
                    if not self.datasetsTabs.widget(i).attributesList[attName].isValueLocked:
                        # print(attName)
                        # print(self.datasetsTabs.widget(i).attributesList[attName].Type)
                        # print(self.datasetsTabs.widget(i).attributesList[attName].Value)
                        print(self.datasetsTabs.widget(i).attributesList[attName].asDict())
                        self.tmpDataset.attrs[ConfigAttribute_Prefix+str(counter)+ConfigAttribute_NameSuffix]  = self.datasetsTabs.widget(i).attributesList[attName].Name
                        self.tmpDataset.attrs[ConfigAttribute_Prefix+str(counter)+ConfigAttribute_ValueSuffix] = self.datasetsTabs.widget(i).attributesList[attName].Value
                        self.tmpDataset.attrs[ConfigAttribute_Prefix+str(counter)+ConfigAttribute_TypeSuffix]  = self.datasetsTabs.widget(i).attributesList[attName].Type
                        self.tmpDataset.attrs[ConfigAttribute_Prefix+str(counter)+ConfigAttribute_ValueLockedSuffix]  = self.datasetsTabs.widget(i).attributesList[attName].isValueLocked
                        self.tmpDataset.attrs[ConfigAttribute_Prefix+str(counter)+ConfigAttribute_LockedSuffix] = self.datasetsTabs.widget(i).attributesList[attName].isLocked
                        counter += 1

                self.tmpDataset.attrs[ConfigOption_Prefix+ConfigOptionName_SelectedChannel]    = self.datasetsTabs.widget(i).channelLine.currentText()
                self.tmpDataset.attrs[ConfigOption_Prefix+ConfigOptionName_ChannelDatatype]    = self.datasetsTabs.widget(i).channelDatatypeLine.currentText()
                # self.tmpDataset.attrs[ConfigOption_Prefix+ConfigOptionName_ChConversionFactor] = self.datasetsTabs.widget(i).channelConversionFactorLine.text()
                # self.tmpDataset.attrs[ConfigOption_Prefix+ConfigOptionName_ChOffset]           = self.datasetsTabs.widget(i).channelOffsetLine.text()

            self.configFileObj.close()
            print("Done saving config file.")
        except Exception as e:
            print("Error: An exception was thrown while trying to save your settings. Do you have sufficient write permissions in your program directory? Exception says: " + str(e))
            if not (self.configFileObj is None):
                self.configFileObj.close()

    def loadSettingsFromFile(self,filename):
        print("")
        print("Loading settings file...")
        try:
            self.configFileObj = h5py.File(filename,"r")
        except Exception as e:
            print("Warning: An exception was thrown while trying to load your settings. This probably is OK if you have not saved any settings before. Exception says: " + str(e))
            return
        try:
            self.dataPathLine.setText(self.configFileObj.attrs[ConfigOptionName_DataPath])
        except Exception as e:
            print("Unable to set data path from configuration file. Exception was thrown. Exception says: " + str(e))
        try:
            self.stationNameLine.setText(self.configFileObj.attrs[ConfigOptionName_StationName])
        except Exception as e:
            print("Unable to set station name from configuration file. Exception was thrown. Exception says: " + str(e))
        try:
            index = self.portLine.findText(self.configFileObj.attrs[ConfigOptionName_PortName], QtCore.Qt.MatchFixedString)
            if index >= 0:
                self.portLine.setCurrentIndex(index)
        except Exception as e:
            print("Unable to set selected device from configuration file. Exception was thrown. Exception says: " + str(e))

        datasetsToLoad = list(self.configFileObj.keys())

        print("Datasets to load: " + str(datasetsToLoad))

        #configuration is stored as datasets
        for datasetName in datasetsToLoad:
            print("")
            print("Loading dataset: " + datasetName)
            attributesList = []
            #read all attributes using .items()
            HDF5ConfigAttributes = list(self.configFileObj[datasetName].attrs.items())
            print(HDF5ConfigAttributes)

            #count available attributes
            #the function simply uses the maximum number of attributes available, as every attribute has a number
            #the numbers start from zero and end at n-1, therefore we start from -1 and then add 1 at the end
            self.tmpAttribCount = -1
            for attr in HDF5ConfigAttributes:
                try:
                    if attr[0][-len(ConfigAttribute_NameSuffix):] == ConfigAttribute_NameSuffix:
                        if attr[0][:len(ConfigAttribute_Prefix)] == ConfigAttribute_Prefix:
                            self.tmp = int(attr[0][len(ConfigAttribute_Prefix):-len(ConfigAttribute_NameSuffix)])
                            if(self.tmp > self.tmpAttribCount):
                                self.tmpAttribCount = self.tmp
                except Exception as e:
                    print("Error while parsing the number of attributes in the config file. Exception says: " + str(e))
            self.tmpAttribCount += 1
            # DONE COUNTING ATTRIBUTES

            for i in range(self.tmpAttribCount):
                attributesList.append(Attribute("", "", "", False, False))

            # read attributes and assign them
            try:
                for attr in HDF5ConfigAttributes:
                    if attr[0][:len(ConfigAttribute_Prefix)] == ConfigAttribute_Prefix:
                        if attr[0][-len(ConfigAttribute_NameSuffix):] == ConfigAttribute_NameSuffix:
                            self.attrNum = int(attr[0][len(ConfigAttribute_Prefix):-len(ConfigAttribute_NameSuffix)])
                            attributesList[self.attrNum].Name = attr[1]
                        elif attr[0][-len(ConfigAttribute_ValueSuffix):] == ConfigAttribute_ValueSuffix:
                            self.attrNum = int(attr[0][len(ConfigAttribute_Prefix):-len(ConfigAttribute_ValueSuffix)])
                            attributesList[self.attrNum].Value = attr[1]
                        elif attr[0][-len(ConfigAttribute_TypeSuffix):] == ConfigAttribute_TypeSuffix:
                            self.attrNum = int(attr[0][len(ConfigAttribute_Prefix):-len(ConfigAttribute_TypeSuffix)])
                            attributesList[self.attrNum].Type = attr[1]
                        elif attr[0][-len(ConfigAttribute_LockedSuffix):] == ConfigAttribute_LockedSuffix:
                            self.attrNum = int(attr[0][len(ConfigAttribute_Prefix):-len(ConfigAttribute_LockedSuffix)])
                            attributesList[self.attrNum].isLocked = attr[1]
                        elif attr[0][-len(ConfigAttribute_ValueLockedSuffix):] == ConfigAttribute_ValueLockedSuffix:
                            self.attrNum = int(attr[0][len(ConfigAttribute_Prefix):-len(ConfigAttribute_ValueLockedSuffix)])
                            attributesList[self.attrNum].isValueLocked = attr[1]

            except Exception as e:
                print("Error while trying to parse data from the configuration file. Exception says: " + str(e))

            print("Loaded attributes list:")
            print("Number of attributes in dataset \"" + datasetName + "\" is: " + str(len(attributesList)))
            [attributesList[i].print() for i in range(len(attributesList))]

            # add dataset and attributes

            #find dataset number within the tabs by name
            if not self.addDatasetWithName(datasetName,False,False):
                print("Unable to add dataset " + datasetName + ". It probably already exists.")


            #find the number of the dataset in the tabs
            self.tmpDatasetNumber = -1
            for j in range(self.datasetsTabs.count()):
                if self.datasetsTabs.tabText(j) == datasetName:
                    self.tmpDatasetNumber = j
                    break


            #if dataset was not found
            if self.tmpDatasetNumber == -1:
                print("Error when trying to read config file. Tab cannot be found after being added or after being found existant already.")
                self.configFileObj.close()
                return

            #add attributes to the list
            # (remember that value locked attributes are not saved, and new value locked attributes are ignored)
            for i in range(len(attributesList)):
                #if the attribute already exists and is locked, just ignore it
                #locked attributes are not supposed to be loaded from the config file
                #this helps in case the status of some attribute was changed from normal to value locked
                try:
                    if self.datasetsTabs.widget(self.tmpDatasetNumber).attributeExists(attributesList[i].Name):
                        if self.datasetsTabs.widget(self.tmpDatasetNumber).attributeValueLocked(attributesList[i].Name):
                            print("Ignoring attribute from config file: " + attributesList[i].Name + ", in dataset: " + datasetName + ", since it is locked.")
                            print(attributesList[i].asDict())
                            continue
                except Exception as e:
                    print("Error while trying to check whether the attribute " + attributesList[i].Name + " Exception says: " + str(e))

                self.datasetsTabs.widget(self.tmpDatasetNumber).addAttribFunc_fixed(attributesList[i], True)

            #load options of the dataset dialog
            for attr in HDF5ConfigAttributes:
                #if the prefix of the attribute in the config file indicates that it's an option
                if attr[0][:len(ConfigOption_Prefix)] == ConfigOption_Prefix:
                    try:
                        if attr[0][-len(ConfigOptionName_SelectedChannel):] == ConfigOptionName_SelectedChannel:
                            index = self.datasetsTabs.widget(self.tmpDatasetNumber).channelLine.findText(attr[1], QtCore.Qt.MatchFixedString)
                            if index >= 0:
                                self.datasetsTabs.widget(self.tmpDatasetNumber).channelLine.setCurrentIndex(index)
                        if attr[0][-len(ConfigOptionName_ChannelDatatype):] == ConfigOptionName_ChannelDatatype:
                            index = self.datasetsTabs.widget(self.tmpDatasetNumber).channelDatatypeLine.findText(attr[1], QtCore.Qt.MatchFixedString)
                            if index >= 0:
                                self.datasetsTabs.widget(self.tmpDatasetNumber).channelDatatypeLine.setCurrentIndex(index)


                    except Exception as e:
                        print("Error while parsing dataset options from config file. Exception says: " + str(e))


        try:
            self.configFileObj.close()
        except:
            print("Error: Could not close the config file. Exception says: " + str(e))


        print("Done loading settings file...")

    def acquisitionThreadError(self, message):
        QMessageBox.information(self,"Error",message)
        self.doStopAcquisition()
        # self.acquirer.wait()

    def createMagneticFieldsList(self, dataToSave, dataCollected):
        def dataset_getter(dsname):
            return dataToSave[dsname]

        def attr_getter(attname: str) -> str:
            parts = attname.split("/")
            if len(parts) == 1:
                raise SyntaxError("Error: Magnetic field equation given with a global attribute. "
                                  "This is not supported in this software!")

            elif len(parts) == 2:
                for i in range(self.datasetsTabs.count()):
                    if self.datasetsTabs.tabText(i) == parts[0]:
                        val = self.datasetsTabs.widget(i).attributesList[parts[1]].Value
                        # check if this is an attribute to be read from data,
                        if re.match(r"([%][A-Z]+[%])", val) is not None:
                            try:
                                val = ApplyAttributeReplacement(val, dataCollected, self.acqConfig, parts[0])
                                self.lastAttrs[attname] = val
                            except Exception as e:
                                print("Error while doing attribute replacement; using the last known value. "
                                      "(this error is expected to happen once for every file write). "
                                      "Error says: " + str(e))
                                try:
                                    val = self.lastAttrs[attname]
                                except Exception as e:
                                    print("Failure in using last value. Replacing by value 1.")
                                    val = re.sub(r"([%][A-Z]+[%])", "1", val)
                        return val

            else:
                print("Error: Attributes should be given in the form: ${dataset/attr}")


        equationString = self.getMagneticFieldEquationString()
        # parsed_eq = gd.parse_main_equation(equationString)
        # units = parsed_eq["units"]
        #index of the dictionary
        units = ""
        dictIndex = "Calculated\nMagnetic\nFields " + (units if units != "[]" else "")
        magFieldsArray = gd.eval_equation(equationString,
                                          attr_getter=attr_getter,
                                          dataset_getter=dataset_getter)
        return {dictIndex: magFieldsArray}

    def processDataFromThreadSlot(self,data,dataCollected):

        self.processDataFromThread(data,dataCollected)
        QCoreApplication.processEvents()

    def processDataFromThread(self,data,dataCollected):
        # Append data to data queue
        # self.dataCollected[data.deviceName].appendData(data)

        if not self.plotLocked:
            self.plotLocked = True

            # print("Points saved: " + str(self.dataCollected[data.deviceName].length()))
            self.skipPlotCounter += 1

            # add the calculated magnetic fields to dataToPlot
            dataToPlot = data.dataToSave.copy()
            magFieldsListDict = self.createMagneticFieldsList(data.dataToSave,dataCollected)
            dataToPlot.update(magFieldsListDict)

            # self.dataAnalysisWidget.appendData(dataToPlot, data.startTime, data.samplingRate)
            self.dataAnalysisWidget.setData(dataToPlot, data.startTime, data.samplingRate)

            # Skip plotting every self.skipPlotEvery
            if self.skipPlotCounter >= self.skipPlotEvery:
                self.dataAnalysisWidget.doPlotData()
                self.skipPlotCounter = 0

            self.plotLocked = False



    #enable start button after the thread has started successfully
    def reenableStartButton(self):
        self.startAcquisitionButton.setDisabled(0)

    def acquisitionTypeChanged(self,index):
        if self.acqTypeOptions[index] == AcquisitionTypeOption_Fake:
            # self.addDatasetButton.setDisabled(True)
            self.refreshDevicesButton.setDisabled(True)
            self.portLine.clear()
            self.portLine.addItem("FakeDevice")
            self.portLine.setDisabled(True)

        elif self.acqTypeOptions[index] == AcquisitionTypeOption_COM:
            # self.addDatasetButton.setDisabled(False)
            self.refreshDevicesButton.setDisabled(False)
            self.refreshDevices()

        else:
            self.addDatasetButton.setDisabled(False)
            QMessageBox.information(self,"Information","Acquisition from files entails reading the files "
            "of the SD Card of the acquisition box and parsing them. Please select the directory where these "
            "files are located. It is highly recommended that you copy these files to your computer and not do "
            "the acquisition from the SD Card directly. The directory must be writable too, so that finished "
            "files are moved to a separate folder.")

            self.acquisitionFromFilesPath = QFileDialog.getExistingDirectory(
                self, "Open Directory", self.acquisitionFromFilesPath, QFileDialog.ShowDirsOnly)
            if self.acquisitionFromFilesPath != "":
                self.refreshDevicesButton.setDisabled(True)
                self.portLine.clear()
                self.portLine.addItem("FromFiles")
                self.portLine.setDisabled(True)
            else:
                self.acquisitionTypeLine.setCurrentIndex(0)

    #what to do when forcing window to close
    def closeEvent(self,event):
        #clean up
        #if thread is created, check that it's stopped
        self.viewLogWidget.close()
        if self.acquirer is not None:
            self.stopAcquisitionSignal.emit()
            waitDialog = QProgressDialog()
            waitDialog.setWindowFlags(Qt.FramelessWindowHint)
            progressBar = QProgressBar()
            cancelButton = QPushButton("")
            waitDialog.setBar(progressBar)
            waitDialog.setCancelButton(cancelButton)
            cancelButton.setVisible(0)
            progressBar.setVisible(0)
            waitDialog.setLabelText("Please wait while the acquisition thread closes... This should not take more than 20 seconds.")
            waitDialog.setMinimum(0)
            waitDialog.setMaximum(0)
            # waitDialog.setValue(0)
            windowIcon = QIcon(os.path.join(ScriptPath, "icon.png"))
            waitDialog.setWindowIcon(windowIcon)
            waitDialog.show()
            QCoreApplication.processEvents()
            self.stopAcquisitionSignal.emit()
            QCoreApplication.processEvents()
            #wait for acquisition thread to "die"
            #no need for this as the thread is a daemon
            # self.acquirer.waitForAcquisitionThread()
            QCoreApplication.processEvents()
            # waitDialog.setValue(100)
            QCoreApplication.processEvents()
            waitDialog.close()

    def cleanUp(self):
        # Clean up everything
        for i in self.__dict__:
            item = self.__dict__[i]
            clean(item)


def GetSoftwareInfo():
    versions = {}
    for name, module in sorted(sys.modules.items()):
        if hasattr(module, '__version__'):
            versions[name] = module.__version__
    return versions

def PrintSoftwareInfo():
    print("Starting GNOME Acquirer v" + SOFTWARE_VERSION + "...")
    print("Imported modules: ")
    print(GetSoftwareInfo())



#!/usr/bin/python3

from enum import Enum
import numpy as np


class AttributeType(Enum):
    String    = 0
    Integer32 = 1
    Integer64 = 2
    Float     = 3
    Double    = 4
    Boolean   = 5


def GetTypesList():
    return list(map(GetTypeNameString, AttributeType))


def GetTypeNameString(attType):
    if attType == AttributeType.String:
        return "String"
    elif attType == AttributeType.Integer32:
        return "Integer32"
    elif attType == AttributeType.Integer64:
        return "Integer64"
    elif attType == AttributeType.Float:
        return "Float32"
    elif attType == AttributeType.Double:
        return "Float64"
    elif attType == AttributeType.Boolean:
        return "Boolean"
    else:
        return "Unknown"


def GetTypeDefaultValue_strType(attTypeStr):
    if attTypeStr == GetTypeNameString(AttributeType.String):
        return ""
    elif attTypeStr == GetTypeNameString(AttributeType.Integer32):
        return "0"
    elif attTypeStr == GetTypeNameString(AttributeType.Integer64):
        return "0"
    elif attTypeStr == GetTypeNameString(AttributeType.Float):
        return "0"
    elif attTypeStr == GetTypeNameString(AttributeType.Double):
        return "0"
    elif attTypeStr == GetTypeNameString(AttributeType.Boolean):
        return "False"
    else:
        return "Unknown"


class Attribute:
    def __init__(self, type, name, value, islocked, isvaluelocked):
        self.Value = value
        self.Type  = type
        self.Name  = name
        self.isLocked = islocked
        self.isValueLocked = isvaluelocked

    def asDict(self):
        return {"value": self.Value,
                "type": self.Type,
                "name": self.Name,
                "isLocked": self.isLocked,
                "isValueLocked": self.isValueLocked}

    def print(self):
        print(self.asDict())


def ApplyAttributeType(attrStr, attType):
    if attType == GetTypeNameString(AttributeType.String):
        return str(attrStr)
    elif attType == GetTypeNameString(AttributeType.Boolean):
        if attrStr.lower() == "false" or attrStr.lower() == "0" or attrStr.lower() == "no":
            return False
        else:
            return True
    elif attType == GetTypeNameString(AttributeType.Float):
        return np.float32(attrStr)
    elif attType == GetTypeNameString(AttributeType.Double):
        return np.float64(attrStr)
    elif attType == GetTypeNameString(AttributeType.Integer32):
        return np.int32(attrStr)
    elif attType == GetTypeNameString(AttributeType.Integer64):
        return np.int64(attrStr)
    else:
        raise ValueError("Bad casting. Cannot cast " + attrStr + " to " + attType)



# from distutils.core import setup
import GNOMEAcquirerLibs.Meta
import os
from setuptools import setup

try:
    del os.link
except:
    pass

includefiles = ["GNOMEAcquirerLibs/icon.png",
                "GNOMEAcquirerLibs/icon-with-debug.ico",
                "GNOMEAcquirerLibs/LogFile.png"]

setup(
    name="gnomeacquirer",
    version = GNOMEAcquirerLibs.Meta.__version__,
    author="Samer Afach",
    author_email="samer@afach.de",
    packages=["GNOMEAcquirerLibs","gnomedata"],
    include_package_data=True,
    url="https://git.afach.de/budker/GNOMEBoxDataAcquisition",
    description="GNOME experiment acquisition software for magnetometers",
    data_files=includefiles,
    #install_requires=['numpy', 'sympy', 'pyserial', 'pyqt', 'h5py', 'matplotlib'],
    python_requires = '>=3.4',
    entry_points={
        'console_scripts': [
            'gnomeacquirer = GNOMEAcquirerLibs.gnomeacquirer:main_function'
        ]},
    scripts=["gnomeacquirer_prepare.bat"]
)



import matplotlib.dates as mdates

import datetime
import copy

#This class is a quick class that could accumulate data from DataBatch to be plotted
class DataToPlot:
    def __init__(self):
        self.dataToSave = {}
        self.MaxSizeToPlot = 10000

    def appendData(self, dataToSave, startTime, samplingRate):
        self.startTime = startTime
        self.samplingRate = samplingRate
        timeArray = []
        timestep = datetime.timedelta(microseconds=float(10 ** 6) / float(samplingRate))
        t0 = copy.deepcopy(startTime)

        # get a random element to get array size
        for self.tmpDSName in dataToSave:
            pass

        # create time array
        for i in range(len(dataToSave[self.tmpDSName])):
            timeArray.append(mdates.date2num(t0))
            t0 += timestep

        if len(dataToSave) == len(self.dataToSave):
            self.dataToPlot_timeAxis = self.dataToPlot_timeAxis + timeArray
            for self.tmpDataset in dataToSave:
                self.dataToSave[self.tmpDataset] = self.dataToSave[self.tmpDataset] + dataToSave[
                    self.tmpDataset].tolist()
            if len(self.dataToSave[self.tmpDataset]) > self.MaxSizeToPlot:
                self.dataToPlot_timeAxis = self.dataToPlot_timeAxis[-self.MaxSizeToPlot:]
                for el in self.dataToSave:
                    self.dataToSave[el] = self.dataToSave[el][-self.MaxSizeToPlot:]
        else:
            self.dataToSave = dataToSave
            for self.tmpDataset in dataToSave:
                self.dataToSave[self.tmpDataset] = dataToSave[self.tmpDataset].tolist()
            self.dataToPlot_timeAxis = timeArray



#!/usr/bin/python3

from PyQt5 import QtCore

from math import sin,cos,pi,floor
import time
import datetime
import copy
import serial
from GNOMEAcquirerLibs.DataCollected import *
from GNOMEAcquirerLibs.DataBatch import *
from GNOMEAcquirerLibs.DataToPlot import *
import os
import sys
import glob
import ctypes as c
import multiprocessing as mp

#constants
stringDumpFolderName  = "Dumped"
stringDumpFilePadding = 6
numOfLinesInDumpFile  = 500000

acquisitionEndlineChar = "\r\n"

NoDevicesString = "Cannot find any device"

def decodeString(strIn):
    if isinstance(strIn,bytes):
        retStr = strIn.decode("ascii")
    else:
        retStr = strIn
    return retStr

def CreateDir(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)

def ListOnlyDirs(path):
    if path[-1] != "/":
        path += "/"
    if os.path.isdir(path):
        contents = os.listdir(path)
    else:
        return []
    contentsOnlyDirs = []
    for p in contents:
        if os.path.isdir(os.path.join(path, p)):
            contentsOnlyDirs.append(p)
    return contentsOnlyDirs


def ListOnlyFiles(path):
    if path[-1] != "/":
        path +=  "/"
    if os.path.isdir(path):
        contents = os.listdir(path)
    else:
        return []
    contentsOnlyFiles = []
    for p in contents:
        if os.path.isfile(os.path.join(path, p)):
            contentsOnlyFiles.append(p)
    return contentsOnlyFiles


def get_serial_ports():
    """ Lists serial port names

        :raises EnvironmentError:
            On unsupported or unknown platforms
        :returns:
            A list of the serial ports available on the system
    """
    if sys.platform.startswith('win'):
        ports = ['COM%s' % (i + 1) for i in range(256)]
    elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
        # this excludes your current terminal "/dev/tty"
        ports = glob.glob('/dev/tty[A-Za-z]*')
    elif sys.platform.startswith('darwin'):
        ports = glob.glob('/dev/tty.*')
    else:
        raise EnvironmentError('Unsupported platform')

    result = []
    for port in ports:
        try:
            s = serial.Serial(port)
            s.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass
    return result

# fills the queue of lines self.acqBuffer
def fillQueue_process(devSpecs, acquisitionSpecs, shared_stopped, shared_return_list, shared_lock):
    def DecodeString(strIn):
        if isinstance(strIn,bytes):
            retStr = strIn.decode("ascii")
        else:
            retStr = strIn
        return retStr

    def TransferQueueToSharedMemory(bufferQueue):
        if len(bufferQueue) > acquisitionSpecs["LengthToPass"]:
            # print("locking in fill")
            try:
                shared_lock.acquire()
                # print("locked in fill")
                if len(shared_return_list) < acquisitionSpecs["AcqMaxBufferLines"]:
                    shared_return_list.extend(bufferQueue)
                    del bufferQueue[:]
                # print("unlocking in fill")
            finally:
                shared_lock.release()
                # print("unlocked in fill")

    def ReopenDevice(devObject):
        try:
            devObject.close()
            devObject.open()
        except Exception as e:
            print("Error while trying to connect to device " + devObject.port + ". The error says: " + str(e))
            time.sleep(2)

    try:
        devObj = serial.Serial()
        devObj.port = devSpecs["DevName"]
        devObj.baudrate = 115200
        devObj.bytesize = serial.EIGHTBITS  # number of bits per bytes
        devObj.parity = serial.PARITY_ODD  # set parity check: no parity
        devObj.stopbits = serial.STOPBITS_ONE  # number of stop bits
        devObj.timeout = 2  # block read for this amount
        devObj.xonxoff = True  # disable software flow control
        devObj.rtscts = False  # disable hardware (RTS/CTS) flow control
        devObj.dsrdtr = False  # disable hardware (DSR/DTR) flow control
        devObj.writeTimeout = 2  # timeout for write

        devObj.open()

        print("Instrument: " + devSpecs["DevName"] + " is open successfully in the process")

    except Exception as e:
        print("Error open serial port: " + str(e))
        print("Error while trying to open instrument: " + devSpecs["DevName"])
        shared_stopped[0] = True
        return


    acqBufferStr = ""
    acqBuffer = []

    #the next line measures how long it has been since the last successful data pull has been made
    #this is to fix a problem with PySerial where it loses its connection to the device some times for no reason
    #the solution is to reopen the port with the function ReopenPort()
    lastAcquisitionTime = datetime.datetime.now()
    deemFailedConnectionAfter = 120 #seconds
    print(acquisitionSpecs)
    if not acquisitionSpecs["ParanoidMode"]:
        # if the local buffer (software buffer) is empty, fill it from the device
        # if len(self.acqBuffer) <= 0:
        while not shared_stopped[0]:
            print("Pulling data from device...")
            # the following variable ensures that if the buffer is empty, then wait, otherwise keep pulling until it's empty
            # keep pulling until the buffer max number of lines is reached (unless the device buffer is empty, there's "break" inside)
            TransferQueueToSharedMemory(acqBuffer)
            while len(acqBuffer) < acquisitionSpecs["AcqMaxBufferLines"] and (not shared_stopped[0]):
                try:
                    # if the user hasn't stopped the acquisition
                    if not shared_stopped[0]:
                        # read available bytes to local buffer
                        buffer = decodeString(devObj.read(8192))

                        #if the buffer is empty, it's possible that connection to the device is lost
                        #if for time that's equal to deemFailedConnectionAfter noting is received, reopen connection
                        if buffer == "":
                            if abs(datetime.datetime.now() - lastAcquisitionTime) > datetime.timedelta(
                                    seconds=deemFailedConnectionAfter):
                                print("Failed in getting data from the device for " + str(
                                    deemFailedConnectionAfter) + " second(s). Trying to reconnect to the device and resetting time counter...")
                                ReopenDevice(devObj)
                                lastAcquisitionTime = datetime.datetime.now()
                            continue

                        else:
                            lastAcquisitionTime = datetime.datetime.now()

                            acqBufferStr += buffer

                            # commented for performance
                            # print("Expected length: " + str(bytesToRead) + "; Received length: " + str(len(buffer)))

                            # find the last endline character in the accumulated string so that the rest can be combined
                            # with the next available bytes
                            lastChar = acqBufferStr.rfind(acquisitionEndlineChar)
                            acqBuffer.extend(acqBufferStr[0:lastChar + 1].split(acquisitionEndlineChar))
                            # delete empty strings from the buffer queue
                            acqBuffer = list(filter(None, acqBuffer))
                            # start the next buffer from the last available endline
                            acqBufferStr = acqBufferStr[lastChar + 1:]
                            # remove \n and \r from the strings
                            acqBuffer = list(map(lambda x: x.replace("\n", "").replace("\r", ""), acqBuffer))

                        # commented for performance
                        # print("Current buffer size (in lines): " + str(len(acqBuffer)))
                    else:
                        break
                except Exception as e:
                    print("Error was thrown while pulling data. The error says: " + str(e))
                    print("Trying to reconnect to the device")
                    ReopenDevice(devObj)

                TransferQueueToSharedMemory(acqBuffer)

    else:
        while not shared_stopped[0]:
            while len(acqBuffer) < acquisitionSpecs["AcqMaxBufferLines"] and (not shared_stopped[0]):
                try:
                    buffer = decodeString(devObj.readline())
                    buffer = buffer.replace(acquisitionEndlineChar, "").replace("\n", "").replace("\r", "")
                    if buffer == "":
                        continue
                    acqBuffer.append(buffer)
                    TransferQueueToSharedMemory(acqBuffer)

                except Exception as e:
                    print("Error was thrown while pulling data. The error says: " + str(e))
                    print("Trying to reconnect to the device")
                    ReopenDevice(devObj)

            TransferQueueToSharedMemory(acqBuffer)

    try:
        devObj.close()
        print("Device " + devSpecs["DevName"] + " closed successfully.")
    except Exception as e:
        print("Unable to close device " + devSpecs["DevName"] + ". Exception says: " + str(e))

    print("Data acquisition thread reached its end successfully.")


class DataAcquirer(QObject):
    postErrorMessage  = QtCore.pyqtSignal(str)
    postResultsSignal = QtCore.pyqtSignal(DataToPlot, DataCollected)
    postNumMissingBatches = QtCore.pyqtSignal(int)
    reenableStartButtonSignal = QtCore.pyqtSignal()
    finishedSignal = QtCore.pyqtSignal()

    def __init__(self,acqConfig):
        # QtCore.QThread.__init__(self)
        QtCore.QObject.__init__(self)

        self.acqConfig = acqConfig

        self.acquisitionThread = None

        self.dataStrQueue = []
        self.stopped = False
        self.openInstrumentObject = None
        self.rm = None
        self.line = None

        self.fakeReaderCounter = -1
        self.fakePhase = 0
        self.fakeCurrentTime = datetime.datetime.utcnow() - datetime.timedelta(days=2 * 36500)
        self.fakeSamplingRate = 100
        self.fakeSamplingRateCounter = 0
        self.fakeChannelOuts = []

        self.fileAcqReader = None
        self.fileAcqListOfFiles = None
        self.fileAcqLastFileRead = None
        self.fileAcqFileLines = []
        self.fileAcqSubFolderForSuccessful = "Done"
        #sleep after how many lines to avoid freezing the GUI (starts with sleeping to enable the stop button)
        self.fileSleepCounterMax = 200000
        self.fileSleepCounter = self.fileSleepCounterMax


        self.acqBuffer = []
        self.acqBufferStr = ""
        self.acqMaxBufferLines = 100000

        # self.dataCollected = {}

        #the following helps in preventing unlimited pulls in a short time
        #if some time passes without data being found, then the program pauses to breathe to prevent overloading the
        #event loop
        self.pullForHowLongBeforeBreathing = 2
        self.pullsCounterLimit = 100
        self.pullBreatheTime   = 3

        self.mpManager = mp.Manager()
        self.shared_return_list = self.mpManager.list()
        self.shared_stopped = self.mpManager.list()
        self.shared_stopped.extend([False])
        self.shared_result_lock = mp.Lock()

    def stopAcquisition(self):
        self.stopped = True
        self.shared_stopped[0] = True

    def emitError(self,message,DoStopAcquisition=True,WithLogFileEntry=True):
        if WithLogFileEntry:
            self.logFileObj.write(str(datetime.datetime.utcnow().strftime(logDateTimeFormat)) + ": " + message + "\n\n")
            self.logFileObj.flush()
            self.logFileObj.flush()
        if DoStopAcquisition:
            self.stopAcquisition()
            self.postErrorMessage.emit(message)

    def healthyWait(self,totalWait):
        _waitTime = 0
        _waitStep = 0.1
        if totalWait < _waitStep:
            time.sleep(totalWait)
        else:
            while _waitTime < totalWait:
                QCoreApplication.processEvents()
                if self.stopped:
                    break
                else:
                    time.sleep(_waitStep)
                    _waitTime += _waitStep


    #fills the queue of lines self.acqBuffer
    def readLineRealToQueue(self):
        if not self.acqConfig.paranoidMode:
            pull_startTime = time.time()

            # if the local buffer (software buffer) is empty, fill it from the device
            if len(self.acqBuffer) <= 0:
                print("Pulling data from device...")

                #the following variable ensures that if the buffer is empty, then wait, otherwise keep pulling until it's empty

                #keep pulling until the buffer max number of lines is reached (unless the device buffer is empty, there's "break" inside)
                while len(self.acqBuffer) < self.acqMaxBufferLines:
                    QCoreApplication.processEvents()
                    try:
                    # if True:

                        #if the user hasn't stopped the acquisition
                        if not self.stopped:

                            #available bytes in device buffer
                            bytesToRead = self.openInstrumentObject.inWaiting()
                            #commented for performance
                            print("Bytes available: " + str(bytesToRead))

                            #if available bytes are zero
                            # the check for zero size is important because the device takes finite amount of time to prepare
                            # the next set of data to send it
                            if bytesToRead == 0:
                                print("Available bytes are zero! Waiting...")
                                # self.healthyWait(1)
                                if abs(time.time() - pull_startTime) > self.pullForHowLongBeforeBreathing:
                                    print("There has been no data to pull from the device for " + str(self.pullForHowLongBeforeBreathing) + " second(s). Breathing...")
                                    self.healthyWait(self.pullBreatheTime)
                                    pull_startTime = time.time()
                                continue

                            else:
                                #data was found, so reset breathing time
                                pull_startTime = time.time()

                                #self.healthyWait(0.01)
                                #read available bytes to local buffer
                                tempBuffer = decodeString(self.openInstrumentObject.read(bytesToRead))
                                self.acqBufferStr += tempBuffer

                                #commented for performance
                                print("Expected length: " + str(bytesToRead) + "; Received length: " + str(len(tempBuffer)))

                                #find the last endline character in the accumulated string so that the rest can be combined
                                #with the next available bytes
                                lastChar = self.acqBufferStr.rfind(acquisitionEndlineChar)
                                self.acqBuffer.extend(self.acqBufferStr[0:lastChar+1].split(acquisitionEndlineChar))
                                #delete empty strings from the buffer queue
                                self.acqBuffer = list(filter(None, self.acqBuffer))
                                #start the next buffer from the last available endline
                                self.acqBufferStr = self.acqBufferStr[lastChar+1:]
                                #remove \n and \r from the strings
                                self.acqBuffer = list(map(lambda x: x.replace("\n", "").replace("\r", ""), self.acqBuffer))

                            #commented for performance
                            print("Current buffer size (in lines): " + str(len(self.acqBuffer)))
                        else:
                            break
                    except Exception as e:
                        print("Assuming acquisition queue is empty... waiting.")
                        print("Exception message: ",e)
                        if len(self.acqBuffer) > 0:
                            pass
                        else:
                            self.healthyWait(5)
                #these lines make acquisition slower
                #print("Received strings:")
                #for line in self.acqBuffer: print(line)
            # lineData = self.acqBuffer.pop(0)

        else:
            QCoreApplication.processEvents()
            if len(self.acqBuffer) <= 0:
                while len(self.acqBuffer) < self.acqMaxBufferLines:
                    QCoreApplication.processEvents()
                    buffer = decodeString(self.openInstrumentObject.readline())
                    buffer = buffer.replace(acquisitionEndlineChar, "")

                    self.acqBuffer.extend(buffer.split(acquisitionEndlineChar))


                # delete empty strings from the buffer queue
                self.acqBuffer = list(filter(None, self.acqBuffer))
                # remove \n and \r from the strings
                self.acqBuffer = list(map(lambda x: x.replace("\n", "").replace("\r", ""), self.acqBuffer))

            # lineData = self.acqBuffer.pop(0)
            # print(lineData)

        # return lineData

    def readLineFile(self):
        QCoreApplication.processEvents()
        self.fileSleepCounter += 1
        # if self.fileSleepCounter >= self.fileSleepCounterMax:
        #     self.fileSleepCounter = 0
        #     print("Breathing...")
        #     self.healthyWait(3)
        #     print("Done breathing.")
        if len(self.fileAcqFileLines) <= 0:
            if len(self.fileAcqListOfFiles) > 0:
                self.fileAcqLastFileRead = self.fileAcqListOfFiles.pop(0)
                with open(os.path.join(self.acqConfig.fileAcqPath,self.fileAcqLastFileRead)) as f:
                    self.fileAcqFileLines = f.readlines()
                    self.fileAcqFileLines = list(map(lambda x: x.replace("\n", "").replace("\r", ""), self.fileAcqFileLines))
                    self.fileAcqFileLines = list(filter(None, self.fileAcqFileLines))
                    f.close()

                return self.fileAcqFileLines.pop(0)
            else:
                self.emitError("All files in the directory given have been read and parsed.")

        else:
            return self.fileAcqFileLines.pop(0)

    def readLineFake(self):
        self.fakeRetVal = ""
        if self.fakeReaderCounter == -1:
            self.fakeCurrentTime += datetime.timedelta(seconds=1)
            self.fakeRetVal = "@Header"
        elif self.fakeReaderCounter == 0:
            self.fakeRetVal = "Date: " + self.fakeCurrentTime.strftime(DataBatch.f_dateFormat)
        elif self.fakeReaderCounter == 1:
            self.fakeRetVal = "Time: " + self.fakeCurrentTime.strftime(DataBatch.f_timeFormat)
        elif self.fakeReaderCounter == 2:
            self.fakeRetVal = "Week number: 1024"
        elif self.fakeReaderCounter == 3:
            self.fakeRetVal = "Time of week: 12345"
        elif self.fakeReaderCounter == 4:
            self.fakeRetVal = "UTC offset: 0  No UTC info!"
        elif self.fakeReaderCounter == 5:
            self.fakeRetVal = "GPS time, GPS PPS, Time from GPS"
        elif self.fakeReaderCounter == 6:
            self.fakeRetVal = "Receiver mode: 7"
        elif self.fakeReaderCounter == 7:
            self.fakeRetVal = "Self survey progress: 100%"
        elif self.fakeReaderCounter == 8:
            self.fakeRetVal = "Warnings: Antenna open, Not tracking satellites, Almanac not complete"
        elif self.fakeReaderCounter == 9:
            self.fakeRetVal = "Decoding status: 0x01 No GPS time"
        elif self.fakeReaderCounter == 10:
            self.fakeRetVal = "Temperature internal [C]: 12.34"
        elif self.fakeReaderCounter == 11:
            self.fakeRetVal = "Temperature external [C]: --"
        elif self.fakeReaderCounter == 12:
            self.fakeRetVal = "Latitude [deg]: 49.9905505"
        elif self.fakeReaderCounter == 13:
            self.fakeRetVal = "Longitude [deg]: 8.2345987"
        elif self.fakeReaderCounter == 14:
            self.fakeRetVal = "Altitude [m]: 198.62"
        elif self.fakeReaderCounter == 15:
            self.fakeRetVal = "@Data"
        elif self.fakeReaderCounter == 16:
            self.fakeRetVal = self.fakeChannelStates
        elif self.fakeReaderCounter == 17:
            if self.fakeSamplingRateCounter < self.fakeSamplingRate:
                self.fakeSamplingRateCounter += 1
                #subtract one from counter to stay here
                self.fakeReaderCounter -= 1
                self.fakeVal = sin(self.fakePhase)
                self.fakePhase += 2*pi*0.3/self.fakeSamplingRate
                self.fakeChannelOutsVals = list(map(lambda x: x*self.fakeVal,self.fakeChannelsAmplitudes))
                self.fakeChannelOuts = []
                for i in range(len(self.fakeChannelsAmplitudes)):
                    if self.fakeChannelsAmplitudes[i] != 0:
                        self.fakeChannelOuts.append(self.fakeChannelOutsVals[i])
                self.fakeRetVal = '  '.join(str(e) for e in self.fakeChannelOuts)
                if self.fakeSamplingRateCounter == self.fakeSamplingRate:
                    self.fakeReaderCounter += 1


        elif self.fakeReaderCounter == 18:
            self.fakeRetVal = "@Magnetic"
        elif self.fakeReaderCounter == 19:
            self.fakeRetVal = "X Y Z off"
        elif self.fakeReaderCounter == 20:
            self.fakeRetVal = "@End"
            self.fakeReaderCounter = -2
            self.fakeSamplingRateCounter = 0
        else:
            self.fakeReaderCounter = -1
            print("Wrong index: " + str(self.fakeReaderCounter))
        self.fakeReaderCounter += 1
        time.sleep(0.001)
        return self.fakeRetVal

    def readLineRealFromQueue(self):
        # self.readLineRealToQueue()
        self.pullBufferFromParallelProcess()
        try:
            lineData = self.acqBuffer.pop(0)
        except Exception as e:
            print("Queue is empty...")
            lineData = ""
        return lineData

    def pullBufferFromParallelProcess(self):
        #copy list from shared buffer and empty it
        # print("locking in read")
        step_size = 10000 #length to copy per step
        if len(self.acqBuffer) < 200000:
            try:
                self.shared_result_lock.acquire()
                # print("locked in read")
                #copy data to local buffer in steps to avoid freezing GUI
                while len(self.shared_return_list) > 0:
                    QCoreApplication.processEvents()
                    if len(self.shared_return_list) < step_size:
                        self.acqBuffer.extend(self.shared_return_list)
                        del self.shared_return_list[:]
                    else:
                        self.acqBuffer.extend(self.shared_return_list[:step_size])
                        del self.shared_return_list[:step_size]
                    QCoreApplication.processEvents()

                # print("unlocking in read")
            finally:
                self.shared_result_lock.release()
                # print("unlocked in read")

        else:
            print("Internal queue is full. No more data will be pulled from the acquisition thread.")
            print("Current size of the internal local queue: " + str(len(self.acqBuffer)))
            print("Current size of the shared queue: " + str(len(self.shared_return_list)))
        QCoreApplication.processEvents()


    def readLine(self):
        QCoreApplication.processEvents()
        if self.acqConfig.isFakeAcquisition:
            line_read = self.readLineFake()
        elif self.acqConfig.isFileAcquisition:
            line_read = self.readLineFile()
        else:
            line_read = self.readLineRealFromQueue()
        return line_read

    def openDumpFile(self):
        if self.acqConfig.doDumpStringsToFile:
            try:
                dumpPath = os.path.join(self.acqConfig.dataPath, stringDumpFolderName)
                CreateDir(dumpPath)
            except Exception as e:
                self.emitError("Couldn't create the path " + dumpPath + ". Make sure you have access to that directory and enough permissions to write. Exception message: " + str(e))
                return False
            self.dumpFilenum = -1
            while True:
                self.dumpFilenum += 1
                self.dumpFilename = os.path.join(dumpPath,str(self.dumpFilenum).zfill(stringDumpFilePadding))
                if not os.path.isfile(self.dumpFilename):
                    try:
                        self.dumpFileObj = open(self.dumpFilename, 'w')
                        self.dumpFileLineCounter = 0
                        self.dumpFileOpen = True
                        return True
                    except Exception as e:
                        self.emitError("Couldn't create the file " + self.dumpFilename + ". Make sure you have access to that directory and enough permissions to write. Exception message: " + str(e))
                        return False


    def writeDumpFileLine(self,line):
        if self.acqConfig.doDumpStringsToFile:
            if not self.dumpFileOpen:
                self.openDumpFile()

            if self.dumpFileLineCounter > numOfLinesInDumpFile:
                self.dumpFileObj.close()
                self.openDumpFile()

            if self.dumpFileOpen:
                self.dumpFileObj.write(line + "\n")
                self.dumpFileObj.flush()
                self.dumpFileLineCounter += 1
                return True
            else:
                return False

    def openLogFile(self,fileName):
        self.logFileObj = open(fileName, 'w')

    def dataError_ClearCollectedData(self,strQueue):
        self.postNumMissingBatches.emit(self.dataCollected.numOfBatches + 1) #+1 for the current batch that has the problem
        self.dataCollected.dumpRemainingDataToFile(strQueue)
        self.dataCollected.clearData()
        self.dataStrQueue.clear()

    def waitForAcquisitionThread(self):
        while True:
            if self.acquisitionThread is not None:
                if not self.acquisitionThread.is_alive():
                    break

                else:
                    QCoreApplication.processEvents()
                    self.healthyWait(1)

    def run(self):
        self.logFilePath = self.acqConfig.logFilePath
        try:
            CreateDir(os.path.join(self.acqConfig.dataPath, logDirName))

            self.openLogFile(self.logFilePath)
        except Exception as e:
            self.emitError("Unable to open acquisition log file. Exception message: " + str(e), True, False)
        except:
            self.emitError("Unable to open acquisition log file. Unknown exception.", True, False)
        # amplitude of channel is channel number
        self.fakeChannelStates = "Ch1 +/-1.25 [V], Ch2  +/-2.5 [V], Ch3 +/-5 [V], Ch4 +/-10 [V]"
        self.fakeChannelsAmplitudes = list(
            map(lambda x: 0 if (x.count("off") > 0) else 1, self.fakeChannelStates.split(",")))
        for i in range(len(self.fakeChannelsAmplitudes)):
            self.fakeChannelsAmplitudes[i] = self.fakeChannelsAmplitudes[i] * float(i + 1)

        print("Opening dump file...")
        self.openDumpFile()
        print("Finished opening dump file.")

        if self.acqConfig.isFakeAcquisition:
            self.openInstrumentObject = ""
            self.dataCollected = DataCollected(self.acqConfig)
            self.readBatchFunction = self.readBatchLineByLine
        elif self.acqConfig.isFileAcquisition:
            self.openInstrumentObject = ""
            self.dataCollected = DataCollected(self.acqConfig)
            self.fileAcqListOfFiles = ListOnlyFiles(self.acqConfig.fileAcqPath)
            self.readBatchFunction = self.readBatchLineByLine
            # self.fileAcqListOfFiles = list(map(lambda x: os.path.join(self.acqConfig.fileAcqPath,x),self.fileAcqListOfFiles))
            try:
                CreateDir(os.path.join(self.acqConfig.fileAcqPath,self.fileAcqSubFolderForSuccessful))
            except Exception as e:
                self.emitError("Unable to create the folder for successful files. Please make sure that the folder where the files are is writable. Exception message: " + str(e))
        else:
            # self.openInstrument(dev)
            self.dataCollected = DataCollected(self.acqConfig)
            self.readBatchFunction = self.readBatchFromDevice

        self.dataToPlot = DataToPlot()

        #start process for data acquisition
        self.healthyWait(2)
        if (not self.acqConfig.isFakeAcquisition) and (not self.acqConfig.isFileAcquisition):
            devSpecs = {}
            devSpecs["DevName"] = self.acqConfig.requestedDevice
            acquisitionSpecs = {}
            acquisitionSpecs["ParanoidMode"] = self.acqConfig.paranoidMode
            acquisitionSpecs["AcqMaxBufferLines"] = self.acqMaxBufferLines
            acquisitionSpecs["LengthToPass"] = 2000
            self.acquisitionThread = mp.Process(target=fillQueue_process, args=(devSpecs, acquisitionSpecs, self.shared_stopped, self.shared_return_list, self.shared_result_lock))
            self.acquisitionThread.daemon = True
            self.acquisitionThread.start()

        self.reenableStartButtonSignal.emit()
        self.acquisitionLoop()

        if (not self.acqConfig.isFakeAcquisition) and (not self.acqConfig.isFileAcquisition):
            self.acquisitionThread.join()

        self.acquisitionThread = None

        # self.closeAllInstruments()
        if not self.logFileObj.closed:
            self.logFileObj.close()

        print("Data acquisition function ended.")
        self.finishedSignal.emit()


    def acquisitionLoop(self):
        LastErrors = []
        while not self.stopped:
            # print("Starting acquisition cycle (batch read)...")

            #read batch from the queue
            self.dataStrQueue.clear()
            pull_startTime = time.time()
            while len(self.dataStrQueue) <= 0:
                self.pullBufferFromParallelProcess()
                self.dataStrQueue = self.readBatchFunction()
                QCoreApplication.processEvents()

                if self.stopped:
                    return

                if len(self.dataStrQueue) <= 0:
                    if abs(time.time() - pull_startTime) > self.pullForHowLongBeforeBreathing:
                        print("There has been no data to pull from the queue for " + str(
                            self.pullForHowLongBeforeBreathing) + " second(s). Breathing...")
                        self.healthyWait(self.pullBreatheTime)
                        pull_startTime = time.time()

                else:
                    pull_startTime = time.time()

            if not self.stopped:
                #parse the received batch
                self.dataB = DataBatch(self.acqConfig, self.dataStrQueue)
                self.batchParseErrorCheck = self.dataB.containsSevereError()
                #deal with errors of the current batch, if they exist
                if len(self.dataB.errorMsgs) > 0:
                    if self.dataB.errorMsgs != LastErrors: #print errors only if there's anything new
                        print("List of errors produced from analyzing data batch (won't be printed again if errors don't change):")
                        [print(err[ERROR_IDENTIFIER_SEVERITY] + ": " + err[ERROR_IDENTIFIER_MESSAGE]) for err in self.dataB.errorMsgs]
                        print("")

                    #set current errors as errors from last time for the next step
                    LastErrors = self.dataB.errorMsgs
                if self.batchParseErrorCheck is None:
                    try:
                        self.dataToPlot.appendData(self.dataB.dataToSave, self.dataB.startTime, self.dataB.samplingRate)
                    except Exception as e1:
                        self.dataError_ClearCollectedData(self.dataStrQueue)
                        self.emitError("A severe error occured in the acquired data: \n\n" + \
                                       "Data failed to parse possibly due to missing chunks. Exception: " + str(e1), False, True)
                    try:
                        self.postResultsSignal.emit(self.dataToPlot, self.dataCollected)
                    except Exception as e1:
                        self.dataError_ClearCollectedData(self.dataStrQueue)
                        self.emitError("Error while posting results to be plotted. Exception says: " + str(e1), False, True)


                    #append parsed batch to DataCollected object
                    appendStatus = self.dataCollected.appendData(self.dataB)
                    if not appendStatus[DataCollected.AppendData_ReturnSuccessKey]:
                        self.dataError_ClearCollectedData(self.dataStrQueue)
                        self.emitError("A severe error occured in the acquired data: \n\n" + appendStatus[DataCollected.AppendData_ReturnErrorMsgKey], False, True)

                    self.dataStrQueue.clear()
                    if appendStatus[DataCollected.AppendData_ReturnIsLastBatchKey]:
                        # move the successful file to the success folder
                        if self.acqConfig.isFileAcquisition:
                            try:
                                successFile = os.path.join(self.acqConfig.fileAcqPath, self.fileAcqLastFileRead)
                                newSuccessFilePath = os.path.join(self.acqConfig.fileAcqPath,
                                                                  self.fileAcqSubFolderForSuccessful,
                                                                  self.fileAcqLastFileRead)
                                os.rename(successFile, newSuccessFilePath)

                            except Exception as e:
                                msg = "Unable to move the succeeded file \"" + successFile + "\" to the new path \"" + newSuccessFilePath + "\". Exception message: " + str(e)
                                print(msg)
                                self.emitError(msg,True,True)

                else:
                    self.dataError_ClearCollectedData(self.dataStrQueue)
                    self.emitError("One or more severe errors were found. "
                                 "The error message says: \n\n" + str(self.batchParseErrorCheck[ERROR_IDENTIFIER_MESSAGE]),False,True)


    def readBatchLineByLine(self):
        QCoreApplication.processEvents()
        dataStrQueue = []
        try:
            # keep reading until the first line is found (@Header)
            while not self.stopped:
                self.line = self.readLine()
                self.writeDumpFileLine(self.line)
                # print(self.line)
                if self.line == DataBatch.d_startString:
                    # print("Acquiring batch, line by line...")
                    # print(self.line)
                    dataStrQueue.append(self.line)
                    break
            # after @Header is found, read lines
            while not self.stopped:
                QCoreApplication.processEvents()
                self.line = self.readLine()
                self.writeDumpFileLine(self.line)
                dataStrQueue.append(self.line)
                if self.line == DataBatch.d_endString:
                    # print(self.line)
                    break
                    # parse data

            return dataStrQueue

        except Exception as e1:
            print("Exception while trying to read. Error: " + str(e1))

    def readBatchFromDevice(self):
        QCoreApplication.processEvents()
        try:
            dataStrQueue = []
            # find the first line of a batch (@Header)
            startIndex = next((i for i in range(len(self.acqBuffer)) if self.acqBuffer[i] == DataBatch.d_startString), None)
            if startIndex is not None:
                endIndex = next((i for i in range(startIndex+1,len(self.acqBuffer)) if self.acqBuffer[i] == DataBatch.d_endString), None)
                if endIndex is not None:
                    dataStrQueue = self.acqBuffer[startIndex:endIndex+1]
                    #delete from the beginning of the buffer to the end, as the beginning is considered corrupt data
                    del self.acqBuffer[0:endIndex+1]

                else:
                    QCoreApplication.processEvents()
            else:
                QCoreApplication.processEvents()

            return dataStrQueue

        except Exception as e1:
            print("Exception while trying to read a batch. Exception says: " + str(e1))

    def openInstrument(self,instrumentName):
        try:
            self.openInstrumentObject = serial.Serial()
            self.openInstrumentObject.port = instrumentName
            self.openInstrumentObject.baudrate = 115200
            self.openInstrumentObject.bytesize = serial.EIGHTBITS #number of bits per bytes
            self.openInstrumentObject.parity   = serial.PARITY_ODD #set parity check: no parity
            self.openInstrumentObject.stopbits = serial.STOPBITS_ONE #number of stop bits
            self.openInstrumentObject.timeout  = 15        #non-block read
            self.openInstrumentObject.xonxoff  = True      #disable software flow control
            self.openInstrumentObject.rtscts   = False     #disable hardware (RTS/CTS) flow control
            self.openInstrumentObject.dsrdtr   = False     #disable hardware (DSR/DTR) flow control
            self.openInstrumentObject.writeTimeout = 2     #timeout for write

            self.openInstrumentObject.open()

            print("Instrument: " + instrumentName + " is open successfully.")
        except Exception as e:
            print("Error open serial port: " + str(e))
            print("Error while trying to open instrument: " + instrumentName)

    def closeInstrument(self):
        if (not self.acqConfig.isFakeAcquisition) and (not self.acqConfig.isFileAcquisition):
            try:
                self.openInstrumentObject.close()

            except Exception as e:
                print("error communicating...: " + str(e))
                print("Error while trying to close instrument: " + self.acqConfig.requestedDevice)

    def closeAllInstruments(self):
        print("Closing open instruments...")
        self.closeInstrument()
        print("Done closing instruments.")

